#include "common.h"
#include "landscapearea.h"
#include <QtDebug>
#include <QPoint>
/*
 * FIXME: All text/comment in this file needs review and may not at this 
 * point be correct. Also, much of the functions need completion and/or
 * combining and/or removal and currenlty exist only as placeholders.
 * 
 This is the "landscape" where we can watch the members of the community behave in their
 environment.
 By "members" I mean the containers for the current generation of genes, and 
 by "environment" I mean the waveform which they are to find the maxima of. 

 The landscape will look like this:

XOFFSET not shown
 +-YLmax------------------------------------------------------landscape (is everything)----+
 |                                                                                         |
 |YEmax->-*                                                                                |
 |       |*                                                                                |
 |<-Xx-->| *                               EnvironmentRect (from XE0,YE0 to XEmax,YEmax)   |
 |       |   *           *                                                                 |
 |       |    *        *  *         *                                                      |
 |       |                 *      ** *   *********************                             |
 |       |     *      *     *  environment[512]               **********                   |
 |       |           *       **  *     *                                ***************    |
 |       |      *   *          **                                                          |
 |       |       ***                                                                       |
 |     Yaxis                                                                               |
 |       |                                                                                 |
 |       |                                                                    |            |
 |       |(XE0,YE0)                                                           v            |
 |       +--------------------------------------------Xaxis--------------------------|     |
 |                         MemberMarkersRect (10 of MMRs)                     Yy     XEmax |
 |(X0,Y0)            (from X0,Y0, to XEmax+1/2(width marker), YE0)                         |
 +-----------------------------------------------------------------------------------------+
                                                                              ^          XLmax
 Where
   (X0, Y0)   are the absolute 0,0 of the landscape display area. Nothing will be displayed 
              outside the box bounded by (X0,Y0),(XLmax,Y0),(XLmax,YLmax),(X0,YLmax);
   (XE0, YE0) are the relative 0,0 of the environment. Waveforms displayed are bounded within
              the box (XE0, YE0), (XEmax),(),(),
   XEmax will be fixed at 511.
   YEmax will be fixed at 255 as well, but this is not tied to the genome as the X axis is.
   XLmax must be at least 1/2-the-width-of-a-member-marker greater than XEmax in order to display
              the member marker at 511.
   Yy is the distance reserved from Y0 to YE0, and must be at least 11 times the-thickness-of-a-
              member-marker. It is expected that the member marker will be 2 bits thick. There will
              be 10 members and a marker to show when the genome contains no variation.
   Xx is just there to move the display away from the edge a bit.
             
 So the landscape is an output only widget, and it must have:
   --the X and Y axis drawn.
   --putMember(int member, int Xval).
   --putAllAgreeMarker(int member, int Xval).
   --drawEnviroment() where the environment is actually a two dimensional array [512][256].
   --clearHistory() where the member marker ghosts are cleared.

 */
#define XOFFSET 10
#define X0 (XOFFSET+10)
#define Xx 10
#define XE0 (X0+Xx)
#define XEmax (XE0+512)
#define XLmax (XEmax+10 )

#define Y0 0
#define Yy 25
#define YE0 (Y0+Yy)
#define YEmax (YE0+400)
#define YLmax (YEmax+10)

//--------------------------------------------------------------------
// Construct it with settings.
//
LandscapeArea::LandscapeArea(QWidget *parent)
    : QWidget(parent)
{
    setPalette(QPalette(QColor(100, 100, 100)));
    setAutoFillBackground(true);
    ptr_to_environment = NULL ;
    ptr_to_community = NULL ;
}
//FIXME: Ugh! Where's the dstructor?

//--------------------------------------------------------------------
// Just the first time, point LandscapeArea to the data.
//
void LandscapeArea::LandscapeArea_setup(int* environment, int (*community)[GENOME_BITS]  )
{
   ptr_to_environment = environment ;
   ptr_to_community = community ;
}

//--------------------------------------------------------------------
//
void LandscapeArea::slot_EnvironmentUpdate(void)
{
   if (ptr_to_environment != NULL) {
      update() ;
   }
}

//--------------------------------------------------------------------
//
void LandscapeArea::slot_LandscapeUpdate(void) //, int* community)
{
   if (ptr_to_environment != NULL) {
      update() ;
      /* From the Qt documentation:
       * "update" clears the widget (usually filling  it with its background color)
       * and sends a paint event to the widget. This results in a call to the paint
       * event function of the widget.
       */
   }
}

//--------------------------------------------------------------------
//
void LandscapeArea::slot_MMRsUpdate(void)
{
   if (ptr_to_community != NULL) {
      update() ;
   }
}

//--------------------------------------------------------------------
//
void LandscapeArea::slot_ClearHistory(void)
{
   if (ptr_to_community != NULL) {
      update() ;
   }
}



//--------------------------------------------------------------------
// This returns a rectangular area for the Environment (the curve).
QRect LandscapeArea::EnvironmentRect() const
{
     QRect result(XE0, -YE0, (XEmax-XE0), -(YEmax-YE0));
     return result;
}


//--------------------------------------------------------------------
// Call this initially, and any time you want to update the
// whole screen.
//
void LandscapeArea::paintLandscape(QPainter &painter)
{
    painter.setPen(Qt::SolidLine);
         // The +10 below just gets the line away from the edge. It
         // is just for looks and is not functional.
    painter.translate(0, YLmax+10);

    // The dark blue box represents the full drawing area of the environment
    // plus the member markers, plus the history markers.
    painter.setPen(Qt::darkBlue);
    painter.drawLine(   Xx,    -Y0, XLmax,    -Y0);
    painter.drawLine(   Xx, -YLmax, XLmax, -YLmax);
    painter.drawLine(   Xx,    -Y0,    Xx, -YLmax);
    painter.drawLine(XLmax,    -Y0, XLmax, -YLmax);
   
    // The cyan box contains only the environment.
    painter.setPen(Qt::cyan);
    painter.drawLine(  XE0,   -YE0, XEmax,   -YE0);
    painter.drawLine(  XE0, -YEmax, XEmax, -YEmax);
    painter.drawLine(  XE0,   -YE0,   XE0, -YEmax);
    painter.drawLine(XEmax,   -YE0, XEmax, -YEmax);
}

//--------------------------------------------------------------------
// Call this to paint (not display) the environment only.
// This function should be drawing only in the same area
// as the rectangle EnvironmentRect.
//
// I am drawing this as a lot of little rectangles so that I can
// choose to make them bigger (2x2 instead of 1x1, for example)
// if aesthetics demand it.
//
void LandscapeArea::paintEnvironment(QPainter &painter)
{
   if (ptr_to_environment != NULL)
   {
      painter.setPen(Qt::NoPen);
      painter.setPen(Qt::gray);
      int x = 0 ;
      int y ;
      for (int i=0;i<ENVIRONMENT_LENGTH;i++)
      {
         y = ptr_to_environment[i] ; 
         //QRect rectangle(XE0+x++,-(YE0+y),1,1); 
         //painter.drawRect(rectangle);
         painter.drawPoint(XE0+x++, -(YE0+y));
      }
   }
}
//--------------------------------------------------------------------
// Call this to paint (not display) the members only.
// This function should be drawing only in the same area
// as the rectangle MemberMarkersRect.
//
void LandscapeArea::paintMemberMarkers(QPainter &painter)
{
   if (ptr_to_community != NULL) {
      painter.setPen(Qt::NoPen);
      painter.setPen(Qt::green);
      int g ;
      for (int m = 0;m<MEMBERS;m++)
      {
         g = genomeAsInteger(GENOME_BITS,ptr_to_community[m]) ;
         if ( g > -1 ) {
            painter.setPen(Qt::black);
            //QRect rectangle( XE0+g, (Y0-2), 1,1);
            painter.drawPoint(XE0+g, (Y0-2));
            //painter.drawRect(rectangle);
            painter.setPen(Qt::green);
         }
         //QRect rectangle( XE0+g, (Y0-4-(2*m)), 2,1);
         //painter.drawRect(rectangle);
         //QPoint point( XE0+g, (Y0-4-(2*m)), 2,1);
         painter.drawPoint(XE0+g-1, (Y0-4-(2*m)));
         painter.drawPoint(XE0+g, (Y0-4-(2*m)));
         painter.drawPoint(XE0+g+1, (Y0-4-(2*m)));
       }
    }
}

//--------------------------------------------------------------------
//Called as a result of the update().
//Why do an update(QRect(thingy)) then?
// 
void LandscapeArea::paintEvent(QPaintEvent * /* event */)
{
   QPainter painter(this);
   paintLandscape(painter);
   paintEnvironment(painter);
   paintMemberMarkers(painter);
}

