#include <QLCDNumber>
#include <QToolTip>
#include <QWidget>
#include <QObject>
#include <QSpinBox>
#include <QTimer>
#include <QDebug>
#include <QFile>
#include <QApplication>
#include <QFont>
#include <QGridLayout>
#include <QHBoxLayout>
#include <QPushButton>
#include <QVBoxLayout>
#include <QComboBox>
#include <QLabel>
#include <QSignalMapper>
#include <QDir>

#include "stdlib.h"

#include "common.h"
#include "lcdrange.h"
#include "mainwindow.h"
#include "landscapearea.h"
#include "gaga.h"

//--------------------------------------------------------------------
MainWindow::~MainWindow()
{
}

//--------------------------------------------------------------------
MainWindow::MainWindow(QWidget *parent) : QWidget(parent)
{
   
   iterations = 10 ; //FIXME: This has been reduced to a flag, finish it. 

   generations = 0 ;
                         
   srand(RANDSEED) ;

   // NOTE:
   // This is probably the only "trick" in the code, setting the timer with
   // with a value of 0. See slot_gaga for details.  
   QTimer *gaga_timer = new QTimer(this);
           gaga_timer_ptr = gaga_timer ;
   connect(gaga_timer, SIGNAL(timeout()), this, SLOT(slot_gaga()));

   // Make a default environment
   //
   for (bit = 0; bit < ENVIRONMENT_LENGTH; bit++)
   {
      if (bit < ENVIRONMENT_LENGTH/2) {
         theEnvironment[bit] = theRefEnvironment[bit] = (int)(bit/2) ;
      }
      else
      {
         theEnvironment[bit] = theRefEnvironment[bit] = (int)(ENVIRONMENT_LENGTH/2 - bit/2) ;
      }
   }

   // Make a default set of genes
   //
   for (membr=0;membr<MEMBERS;membr++) {
      for (bit=0 ; bit<GENOME_BITS;bit++) {
         theCommunity[membr][bit] = rand() % 2; } }


   ttip_clearhistBTN = "Currently DISABLED" ;
   ttip_fitnessBOX   = "The fitness value 0-99 is simply passed\nuntouched to the GA." ;
   ttip_mutebiasBOX  = "Logarithmically prefer mutations on the LSBs.\n0 means no bias,\n1 means a weak bias towards small changes,\n10 means mutations to the MSB are very rare.";
   ttip_muteprobBOX  = "Change the probability of a\nmutation (bit flip) in a random\nmember's genome at a random place.";

   // /////////////////////////////////////////////////////////////////////////////
   // Now Create all the GUI stuff
   //
   setWindowTitle(tr("gaga-ninebit")) ;

   // ----------------------------------------------------------------
   // This is the drawing area for all the GA activity, the environment plus
   // the markers for the community members. Both the members and the environment
   // are contained within the "landscape".
   LandscapeArea *Lbox = new LandscapeArea;
      Lbox_ptr = Lbox ;
      Lbox->LandscapeArea_setup(theEnvironment, theCommunity) ;

   // ----------------------------------------------------------------
   QPushButton *quitBTN = new QPushButton(tr("&Quit"));
                quitBTN->setFont(QFont("Serif", 16, QFont::Bold));
        connect(quitBTN, SIGNAL(clicked()), qApp, SLOT(quit()));


   // The top row of Buttons and stuff ...
   // ----------------------------------------------------------------
   QPushButton *loadBTN = new QPushButton(tr("Load File:"));
                loadBTN_ptr = loadBTN ;
                loadBTN->setFont(QFont("Serif", 14, QFont::Bold));
        connect(loadBTN, SIGNAL(clicked()), this, SLOT(slot_LoadFile())) ; 

   QComboBox *InputFileComboBox = new QComboBox;
      InputFileComboBox_ptr = InputFileComboBox ;
      InputFileComboBox->setEditable(true);
      InputFileComboBox->addItem(tr("anicedampedsinusoid"));
      InputFileComboBox->addItem(tr("randomwalk"));
      InputFileComboBox->addItem(tr("twosimilarpeaks"));
      InputFileComboBox->addItem(tr("twodissimilarpeaks"));
      InputFileComboBox->addItem(tr("user"));
      InputFileComboBox->setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Preferred);

   QPushButton *stopBTN   = new QPushButton(tr("Stop")); 
                stopBTN_ptr = stopBTN ;
                stopBTN->setFont(QFont("Serif", 14, QFont::Bold));
        connect(stopBTN, SIGNAL(clicked()), this, SLOT(slot_StopState()));

   QPushButton *runBTN    = new QPushButton(tr("Run")); 
                runBTN_ptr = runBTN ;
                runBTN->setFont(QFont("Serif", 14, QFont::Bold));
        connect(runBTN, SIGNAL(clicked()),               this, SLOT(slot_RunPreState()));

   QPushButton *stepBTN    = new QPushButton(tr("Step")); 
                stepBTN_ptr = stepBTN ;
                stepBTN->setFont(QFont("Serif", 14, QFont::Bold));
        connect(stepBTN, SIGNAL(clicked()),              this, SLOT(slot_StepPreState()));
        connect(this, SIGNAL(signal_RunPreState(int)),   this, SLOT(slot_RunState(int)));

   QLabel *generationLabel = new QLabel(tr("Generations:"));
   QLCDNumber *generationLCD    = new QLCDNumber(2); 
               generationLCD_ptr = generationLCD ;
               generationLCD->setSegmentStyle(QLCDNumber::Flat) ;
               generationLCD->setNumDigits(8) ;
               generationLCD->display(generations) ;


   // The second row of Buttons and stuff ...
   // ----------------------------------------------------------------
   QPushButton *reloadBTN = new QPushButton(tr("Reload")); 
                reloadBTN->setFont(QFont("Serif", 14, QFont::Bold));
        connect(reloadBTN, SIGNAL(clicked()), this, SLOT(slot_Reload()));

   QPushButton *clearhistBTN = new QPushButton(tr("Clear History")); 
                clearhistBTN->setFont(QFont("Serif", 14, QFont::Bold));
                clearhistBTN->setToolTip(ttip_clearhistBTN);
        connect(clearhistBTN, SIGNAL(clicked()), Lbox_ptr, SLOT(slot_ClearHistory()));

   QPushButton *LShiftBTN = new QPushButton(tr("shift Left"));
                LShiftBTN->setFont(QFont("Serif", 14, QFont::Bold));
                LShiftBTN->setAutoRepeat(TRUE);
                LShiftBTN->setAutoRepeatInterval(10);
        connect(LShiftBTN, SIGNAL(clicked()), this, SLOT(slot_LShift()));

   QPushButton *RShiftBTN = new QPushButton(tr("shift Right"));
                RShiftBTN->setFont(QFont("Serif", 14, QFont::Bold));
                RShiftBTN->setAutoRepeat(TRUE);
                RShiftBTN->setAutoRepeatInterval(10);
        connect(RShiftBTN, SIGNAL(clicked()), this, SLOT(slot_RShift()));

   QPushButton *UShiftBTN = new QPushButton(tr("shift Up"));
                UShiftBTN->setFont(QFont("Serif", 14, QFont::Bold));
                UShiftBTN->setAutoRepeat(TRUE);
                UShiftBTN->setAutoRepeatInterval(10);
        connect(UShiftBTN, SIGNAL(clicked()), this, SLOT(slot_UShift()));

   QPushButton *DShiftBTN = new QPushButton(tr("shift Down"));
                DShiftBTN->setFont(QFont("Serif", 14, QFont::Bold));
                DShiftBTN->setAutoRepeat(TRUE);
                DShiftBTN->setAutoRepeatInterval(10);
        connect(DShiftBTN, SIGNAL(clicked()), this, SLOT(slot_DShift()));


   // The left side column of control Buttons and stuff ...
   // ----------------------------------------------------------------
   QPushButton *newgenesBTN = new QPushButton(tr("New Genes"));
                newgenesBTN_ptr = newgenesBTN ;
                newgenesBTN->setFont(QFont("Serif", 14, QFont::Bold));
        connect(newgenesBTN, SIGNAL(clicked()), this, SLOT(slot_NewGenes()));

   LCDRange *fitnessBOX = new LCDRange(tr("Fitness Multiplier"));
             fitnessBOX->setToolTip(ttip_fitnessBOX);
             fitnessBOX->setRange(1, 100);
     connect(fitnessBOX, SIGNAL(valueChanged(int)),   this,       SLOT(setFitness(int)));
     connect(this,       SIGNAL(FitnessChanged(int)), fitnessBOX, SLOT(setValue(int)));
     fitnessBOX->setValue(20);

   LCDRange *muteprobBOX = new LCDRange(tr("% Prob(Mutation)"));
             muteprobBOX->setToolTip(ttip_muteprobBOX);
             muteprobBOX->setRange(0, 99);
     connect(muteprobBOX, SIGNAL(valueChanged(int)),   this,         SLOT(setMuteProb(int)));
     connect(this,        SIGNAL(MuteProbChanged(int)), muteprobBOX, SLOT(setValue(int)));
     muteprobBOX->setValue(25); 

   LCDRange *mutebiasBOX = new LCDRange(tr("Mutation Bias"));
             mutebiasBOX->setToolTip(ttip_mutebiasBOX);
             mutebiasBOX->setRange(0, 10);
     connect(mutebiasBOX, SIGNAL(valueChanged(int)),      this,      SLOT(setMuteBias(int)));
     connect(this,        SIGNAL(MuteBiasChanged(int)), mutebiasBOX, SLOT(setValue(int)));
     mutebiasBOX->setValue(1); 


   QLabel *ParamLabel = new QLabel(tr("Additional Parameters\n" "Range %1 to %2:").arg(-100).arg(100));
   QLabel *integerLabelParam0 = new QLabel(tr("Parameter 0:"));
           integerLabelParam0_ptr = integerLabelParam0 ;
   QLabel *integerLabelParam1 = new QLabel(tr("Parameter 1:"));
           integerLabelParam1_ptr = integerLabelParam1 ;
   QLabel *integerLabelParam2 = new QLabel(tr("Parameter 2:"));
           integerLabelParam2_ptr = integerLabelParam2 ;
   QLabel *integerLabelParam3 = new QLabel(tr("Parameter 3:"));
           integerLabelParam3_ptr = integerLabelParam3 ;

   QSpinBox *integerSpinBoxParam0 = new QSpinBox;
             integerSpinBoxParam0_ptr = integerSpinBoxParam0 ;
             integerSpinBoxParam0->setRange(-100, 100);
             integerSpinBoxParam0->setSingleStep(1);
             integerSpinBoxParam0->setValue(0);
   QSpinBox *integerSpinBoxParam1 = new QSpinBox;
             integerSpinBoxParam1_ptr = integerSpinBoxParam1 ;
             integerSpinBoxParam1->setRange(-100, 100);
             integerSpinBoxParam1->setSingleStep(1);
             integerSpinBoxParam1->setValue(0);
   QSpinBox *integerSpinBoxParam2 = new QSpinBox;
             integerSpinBoxParam2_ptr = integerSpinBoxParam2 ;
             integerSpinBoxParam2->setRange(-100, 100);
             integerSpinBoxParam2->setSingleStep(1);
             integerSpinBoxParam2->setValue(0);
   QSpinBox *integerSpinBoxParam3 = new QSpinBox;
             integerSpinBoxParam3_ptr = integerSpinBoxParam3 ;
             integerSpinBoxParam3->setRange(-100, 100);
             integerSpinBoxParam3->setSingleStep(1);
             integerSpinBoxParam3->setValue(0);



   // ----------------------------------------------------------------
   // Place the top layout
   QHBoxLayout *topLayout = new QHBoxLayout;
   topLayout->addWidget(loadBTN);
   topLayout->addWidget(InputFileComboBox);
   topLayout->addWidget(stopBTN);
   topLayout->addWidget(runBTN);
   topLayout->addWidget(stepBTN);
   topLayout->addWidget(generationLabel);
   topLayout->addWidget(generationLCD);
   topLayout->addStretch(1);

   // ----------------------------------------------------------------
   // Place just beneath the top layout
   QHBoxLayout *sousTopLayout = new QHBoxLayout;
   sousTopLayout->addWidget(reloadBTN);
   sousTopLayout->addWidget(clearhistBTN);
   sousTopLayout->addWidget(LShiftBTN);
   sousTopLayout->addWidget(RShiftBTN);
   sousTopLayout->addWidget(UShiftBTN);
   sousTopLayout->addWidget(DShiftBTN);
   sousTopLayout->addStretch(1);

   // ----------------------------------------------------------------
   // Place the left layout
   QVBoxLayout *leftLayout = new QVBoxLayout;
   leftLayout->addWidget(newgenesBTN);
   leftLayout->addWidget(fitnessBOX);
   leftLayout->addWidget(muteprobBOX);
   leftLayout->addWidget(mutebiasBOX);
   leftLayout->addWidget(ParamLabel);
   leftLayout->addWidget(integerLabelParam0);
   leftLayout->addWidget(integerSpinBoxParam0);
   leftLayout->addWidget(integerLabelParam1);
   leftLayout->addWidget(integerSpinBoxParam1);
   leftLayout->addWidget(integerLabelParam2);
   leftLayout->addWidget(integerSpinBoxParam2);
   leftLayout->addWidget(integerLabelParam3);
   leftLayout->addWidget(integerSpinBoxParam3);
   leftLayout->addStretch(1);


   // ----------------------------------------------------------------
   // Put all layout parts together ...
   QGridLayout *gridLayout = new QGridLayout;
   gridLayout->addWidget(quitBTN,       0, 0);
   gridLayout->addLayout(topLayout,     0, 1);
   gridLayout->addLayout(sousTopLayout, 1, 1);
   gridLayout->addLayout(leftLayout,    2, 0);
   gridLayout->addWidget(Lbox,          2, 1);
   gridLayout->setColumnStretch(1, 10) ;
   setLayout(gridLayout);

   // ----------------------------------------------------------------
   // Signal and slot connections NOT grouped with their widgets (above).
   connect(this, SIGNAL(signal_StopState()),
           this,   SLOT(  slot_StopState()));

   // Events that can cause a change in theEnvironment[]:
   // ... entering the Run state:
   connect(this,   SIGNAL(signal_LandscapeUpdate()),
           Lbox_ptr, SLOT(  slot_LandscapeUpdate()));
   // ... changes in amplitude, shift, load or reload:
   connect(this,   SIGNAL(signal_EnvironmentUpdate()),
           Lbox_ptr, SLOT(  slot_EnvironmentUpdate()));

   // Events that can cause a change in the members genes:
   // ... entering the Run state:
   connect(this,   SIGNAL(signal_MMRsUpdate()),  /*From slot_gaga or slot_RunState*/
           Lbox_ptr,   SLOT(slot_MMRsUpdate()));
   // ... NewGenes
   connect(this,   SIGNAL(signal_GenomeUpdate()), /*This signal is from user input*/
           Lbox_ptr,     SLOT(slot_MMRsUpdate()));

   // NOTE:
   // The other Buttons and Sliders, Mutation Bias, Mutation Probability, and
   // Fitness Multiplier, only change the parameters of gaga, and
   // may not have on screen consequences immediately.

   // Start Stopped.
   emit signal_StopState() ;
   // All behaviour will now be in response to user actions.
}

// End of MainWindow::MainWindow

//=========================================================================
// Slots just resulting in parameter changes, not necessarily immediately
// apparent screen updates:
//------------------------------------------------------
void MainWindow::setFitness(int fitness)
{
     if (fitness <= 1)              fitness = 1;
     if (fitness > 99)              fitness = 99;
     if (currentFitness == fitness) return;
     currentFitness = fitness;
     emit FitnessChanged(currentFitness);
}

//------------------------------------------------------
//FIXME: Not yet implented, just a placeholder:
void MainWindow::setAmplitude(int amplitude)
{
     if (amplitude < 1)               amplitude = 1;
     if (amplitude > 99)             amplitude = 99;
     if (currentAmplitude == amplitude) return;
     currentAmplitude = amplitude;
     emit AmplitudeChanged(currentAmplitude);
}

//------------------------------------------------------
void MainWindow::setMuteProb(int muteprob)
{
     if (muteprob < 0)              muteprob = 0;
     if (muteprob > 99)             muteprob = 99;
     if (currentMuteProb == muteprob) return;
     currentMuteProb = muteprob;
     emit MuteProbChanged(currentMuteProb);
}

//------------------------------------------------------
void MainWindow::setMuteBias(int mutebias)
{
     if (mutebias < 0)              mutebias = 0;
     if (mutebias > 10)             mutebias = 10;
     if (currentMuteBias == mutebias) return;
     currentMuteBias = mutebias;
     emit MuteBiasChanged(currentMuteBias);
}



//=========================================================================
// Slots with signals to cause immediate screen updates to the MMRs space:
//
void MainWindow::slot_NewGenes(void) // new RANDOM genes
{
   for (membr=0;membr<MEMBERS;membr++) {
      for (bit=0 ; bit<GENOME_BITS;bit++) {
         theCommunity[membr][bit] = rand() % 2;
     }
   }
   emit signal_GenomeUpdate() ;
}

//=========================================================================
// Slots with signals to cause immediate screen updates to the environment space:
//------------------------------------------------------
void MainWindow::slot_Reload(void)
{
   for (int bit = 0; bit < ENVIRONMENT_LENGTH - 1; bit++)
   {
      theEnvironment[bit] = theRefEnvironment[bit] ;
   }
   emit signal_EnvironmentUpdate() ;
}

//------------------------------------------------------
void MainWindow::slot_LShift(void)
{
   int temp ;
   for (int bit = 0; bit < ENVIRONMENT_LENGTH - 1; bit++) {
      if (bit == 0) { temp = theEnvironment[bit] ; }
      theEnvironment[bit] = theEnvironment[bit+1] ; 
   }
   theEnvironment[ENVIRONMENT_LENGTH - 1] = temp; 
   emit signal_EnvironmentUpdate() ;
}

//------------------------------------------------------
void MainWindow::slot_RShift(void)
{
   int temp ;
   for (int bit = (ENVIRONMENT_LENGTH-1) ; bit > 0 ; bit--) {
      if (bit == (ENVIRONMENT_LENGTH-1)) { temp = theEnvironment[bit] ; }
      theEnvironment[bit] = theEnvironment[bit-1] ; 
   }
   theEnvironment[0] = temp; 
   emit signal_EnvironmentUpdate() ;
}

//------------------------------------------------------
void MainWindow::slot_UShift(void)
{
   for (int bit = 0; bit < ENVIRONMENT_LENGTH ; bit++) {
      theEnvironment[bit] += 1 ; 
   }
   emit signal_EnvironmentUpdate() ;
}

//------------------------------------------------------
void MainWindow::slot_DShift(void)
{
   for (int bit = 0; bit < ENVIRONMENT_LENGTH ; bit++) {
      if (theEnvironment[bit] <= 1 ) {
         theEnvironment[bit] = 1 ; // NOTE: See gaga.cpp, but gaga lives in
                                   // a totally positive integer space. Negative
                                   // integers are not good, and Zero is
                                   // particularly bad.
      }
      else {
         theEnvironment[bit] -= 1 ; 
      }
   }
   emit signal_EnvironmentUpdate() ;
}

//---------------------------------------------------------------------------------
//
void MainWindow::slot_LoadFile() // QString afilename) 
{
   QString afilename = QDir().currentPath() + '/' + InputFileComboBox_ptr->currentText() ; 
   //qDebug() << "afilename is " << afilename << endl ;
   QFile environmentFile(afilename) ;
   environmentFile.open(QFile::ReadOnly);
   QTextStream inputStream(&environmentFile);

   int eel = 0 ;
   while ((!inputStream.atEnd()) && (eel<ENVIRONMENT_LENGTH)) {
      QString word;
      inputStream >> word;
      if (!word.isEmpty()) {
         theEnvironment[eel++] =  atoi((const char*)word.toLatin1().data()) ;
      }
   }
   emit signal_EnvironmentUpdate() ;
}

//------------------------------------------------------
// Just from the Run button
void MainWindow::slot_RunPreState(void) 
{
   // This state is to escape from unnecessary button
   // flicker for a Step.
            newgenesBTN_ptr->setEnabled(FALSE);
                loadBTN_ptr->setEnabled(FALSE);
                 runBTN_ptr->setEnabled(FALSE);
                stopBTN_ptr->setEnabled(  TRUE  );
     integerLabelParam0_ptr->setEnabled(FALSE);
     integerLabelParam1_ptr->setEnabled(FALSE);
     integerLabelParam2_ptr->setEnabled(FALSE);
     integerLabelParam3_ptr->setEnabled(FALSE);
   integerSpinBoxParam0_ptr->setEnabled(FALSE);
   integerSpinBoxParam1_ptr->setEnabled(FALSE);
   integerSpinBoxParam2_ptr->setEnabled(FALSE);
   integerSpinBoxParam3_ptr->setEnabled(FALSE);

   emit signal_RunPreState(iterations) ;
}

//------------------------------------------------------
void MainWindow::slot_StepPreState(void) 
{
   emit signal_RunPreState(1) ;
}

//------------------------------------------------------
void MainWindow::slot_gaga(void) 
{
   int return_val = gaga(
         currentMuteProb,
         currentMuteBias,
         currentFitness,
         theEnvironment,
         theCommunity,
         Param0,
         Param1,
         Param2,
         Param3
         ) ;
      // If gaga returned, signal_MMRsUpdate was emitted to update the display,
      // and gaga was called immediately, the display would not update until
      // someone stopped calling gaga. The event queue just queues up the screen
      // events to process at its leisure and the gaga routine takes precedence.
      // The gaga_timer_ptr->start() and stop are used to avoid this problem 
      // (because it is no fun unless you can watch the members update each 
      // generation. According to the Qt documentation:
      //     As a special case, a QTimer with a timeout of 0 will
      //     time out as soon as all the events in the window
      //     system's event queue have been processed.
      //generations++ ;
      generationLCD_ptr->display(++generations) ;
      emit signal_MMRsUpdate() ;
      // So the screen update signal above queues up the event, but with the
      // timer started gaga will not be called again until after that update.
}

//------------------------------------------------------
void MainWindow::slot_RunState(int steps)
{
   static int firsttimethrough = 1 ; 
   if (firsttimethrough == 1)
   {
      firsttimethrough = 0 ; 
      emit signal_LandscapeUpdate(); 
      emit signal_EnvironmentUpdate() ;
      emit signal_MMRsUpdate() ;
   }
   if (steps == 1) {
      int ret_val = gaga(
           currentMuteProb,
           currentMuteBias,
           currentFitness,
           theEnvironment,
           theCommunity,
           Param0,
           Param1,
           Param2,
           Param3
           ) ; 
      //generations++ ;
      generationLCD_ptr->display(++generations) ;
      emit signal_MMRsUpdate() ;
   }
   else {
      gaga_timer_ptr->start();
   }
}


//---------------------------------------------------------------------------------
//
void MainWindow::slot_StopState()
{
   gaga_timer_ptr->stop();
            newgenesBTN_ptr->setEnabled(TRUE);
                loadBTN_ptr->setEnabled(TRUE);
                 runBTN_ptr->setEnabled(TRUE);
                stopBTN_ptr->setEnabled(  FALSE  );
     integerLabelParam0_ptr->setEnabled(TRUE);
     integerLabelParam1_ptr->setEnabled(TRUE);
     integerLabelParam2_ptr->setEnabled(TRUE);
     integerLabelParam3_ptr->setEnabled(TRUE);
   integerSpinBoxParam0_ptr->setEnabled(TRUE);
   integerSpinBoxParam1_ptr->setEnabled(TRUE);
   integerSpinBoxParam2_ptr->setEnabled(TRUE);
   integerSpinBoxParam3_ptr->setEnabled(TRUE);
}

